/******************************************************************************
 * mesh.h 
 * implements the mesh data structure
 * Copyright (C) 2010 Marc Gissler <gisslerm@informatik.uni-freiburg.de>
 ******************************************************************************/

#ifndef MESHES_H
#define MESHES_H

#include "globals.h"

#include <string>
#include <vector>

#include "gmMath\gmVec3.h"
#include "gmMath\gmVec4.h"
#include "gmMath\gmMat4.h"

// ------------------------------------------------------------
// defines thes mesh class
// 

class CMesh {

public:
	// misc stuff

	typedef gmVector3 Vertex;

	struct Edge
	{
		int v0Id, v1Id;
		int neightborFace[2];

		// constructors
		Edge() : v0Id(-1), v1Id(-1) 
		{ 
			neightborFace[0] = -1;
			neightborFace[1] = -1;
		}
		Edge(int _v0Id, int _v1Id) : v0Id(_v0Id), v1Id(_v1Id) 
		{}

		// comparison functions and operators

		static void sort(Edge& e)
		{
			if(e.v0Id > e.v1Id) std::swap(e.v0Id, e.v1Id);
		}

		friend bool operator<(const Edge& _e0, const Edge& _e1)
		{
			Edge e0(_e0);
			Edge e1(_e1);

			sort(e0);
			sort(e1);

			if(e0.v0Id < e1.v0Id) return true;
			if(e0.v0Id > e1.v0Id) return false;
			if(e0.v1Id < e1.v1Id) return true;
			return false;
		}

		friend bool operator==(const Edge& _e0, const Edge& _e1)
		{
			Edge e0(_e0);
			Edge e1(_e1);

			sort(e0);
			sort(e1);

			if(e0.v0Id != e1.v1Id) return false;
			if(e0.v1Id != e1.v1Id) return false;
			return true;
		}
	};

	typedef gmVector3 Normal;

	struct Face 
	{
		int v1, v2, v3;	// vertex IDs
		int n1, n2, n3;	// normal IDs
		int t1, t2, t3;	// texture IDs
		gmVector3 normal; // vector orthogonal to the plane the face lies in
		bool bLit;
	};

	struct TextureCoordinate
	{
		float u, v;
	};

	struct Texture 
	{
		int width, height;
		void *pTextureData;
	};

	// member variables
private:
	std::vector<Vertex>					m_vectorVertices;
	std::vector<Edge>						m_vectorEdges;
	std::vector<Face>						m_vectorFaces;	
	std::vector<Normal>					m_vectorNormals;
	std::vector<TextureCoordinate>	m_vectorTextureCoordinates;

	GLfloat m_ambientColor[4];
	GLfloat m_diffuseColor[4];
	GLfloat m_specularColor[4];

public:
	gmMatrix4 m_transform;

public:
	//constructor/destructor
	CMesh(std::string& filename);
	~CMesh();

	// adds a rotation to the mesh
	void addRotation(double angle, gmVector3& _axis);

	// adds a translation to the mesh
	void addTranslation(gmVector3& _translation);

	// clears the mesh
	void clear();

	// draws the mesh
	void draw() const;

	const std::vector<Edge>& getEdges() const;
	std::vector<Face>& getFaces();
	const std::vector<Vertex>&	getVertices()const ;
	const std::vector<gmVector3>&	getNormals() const;
	const std::vector<TextureCoordinate>& getTextureCoordinates() const;

	// loads a mesh from an obj file
	void loadObjFile(std::string& filename);
	
	// creates a diffuse material
	void setDiffuseMaterial(GLfloat red, GLfloat green, GLfloat blue);
	
	// creates a specular material
	void setSpecularMaterial(GLfloat red, GLfloat green, GLfloat blue);
};

#endif

