/******************************************************************************
 * shadowmap.cpp 
 * implements the construction and drawing of shadow maps
 * Copyright (C) 2010 Marc Gissler <gisslerm@informatik.uni-freiburg.de>
 ******************************************************************************/

#include "shadowmap.h"

#include "shaderProgramDepth.h"
#include "shaderProgramShadowMap.h"

#include "lightmanager.h"
#include "scene.h"
#include "trackball/glTrackball.h"

CShadowMap::CShadowMap() : m_shaderShadowMap(NULL), m_shaderDepth(NULL)
{
	m_shadowMapSize[0] = 64;
	m_shadowMapSize[1] = 64;

  // prepare shadow map textures
	glGenTextures(1, &m_shadowMapDepth);
	glBindTexture(GL_TEXTURE_2D, m_shadowMapDepth);
	glTexParameteri(GL_TEXTURE_2D, GL_DEPTH_TEXTURE_MODE, GL_LUMINANCE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_R_TO_TEXTURE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC, GL_LEQUAL);
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP/*GL_CLAMP_TO_EDGE*/);
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP/*GL_CLAMP_TO_EDGE*/);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, m_shadowMapSize[0], m_shadowMapSize[1], 0, GL_DEPTH_COMPONENT, GL_FLOAT, NULL);

	glGenTextures(1, &m_shadowMapColor);
	glBindTexture(GL_TEXTURE_2D, m_shadowMapColor);
	glTexImage2D(GL_TEXTURE_2D, 0, /*GL_RGBA*/GL_RGBA32F_ARB,  m_shadowMapSize[0], m_shadowMapSize[1], 0, GL_RGB, GL_FLOAT, NULL);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP/*GL_CLAMP_TO_EDGE*/);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP/*GL_CLAMP_TO_EDGE*/);

	//create frame buffer
	glGenFramebuffersEXT(1, &m_shadowFrameBufferHandle);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_shadowFrameBufferHandle);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_TEXTURE_2D, m_shadowMapDepth, 0);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, m_shadowMapColor, 0);

	// if the frame buffers could not be created, exit the program
	if(!m_checkValidity()) 
	{
		glDeleteTextures(1, &m_shadowMapDepth);			
		glDeleteTextures(1, &m_shadowMapColor);			
		glDeleteFramebuffersEXT(1,  &m_shadowFrameBufferHandle);
		m_shadowMapDepth = m_shadowMapColor = m_shadowFrameBufferHandle = 0;
		exit(-1);
	}
	
	// unbind, for now (to get back to the original frame buffer)
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
}

CShadowMap::~CShadowMap()
{
	glDeleteTextures(1, &m_shadowMapDepth);
	glDeleteTextures(1, &m_shadowMapColor);

	glDeleteFramebuffersEXT(1,  &m_shadowFrameBufferHandle);
}

void CShadowMap::build(int lightId, gmVector3& lightPos)
{
	std::map<int, CShadowData*>::iterator mapIter;
	mapIter = m_mapLightAndData.find(lightId);

	if(mapIter == m_mapLightAndData.end())
	{
		CShadowData* data = new CShadowData();
		mapIter = m_mapLightAndData.insert(std::pair<int, CShadowData*>(lightId, data)).first;
	}
}

void CShadowMap::clear()
{
	std::map<int, CShadowData*>::iterator mapIter;
	for(mapIter = m_mapLightAndData.begin(); mapIter != m_mapLightAndData.end(); ++mapIter)
	{
		CShadowData* data = mapIter->second;
		delete data;
	}
	m_mapLightAndData.clear();
}

bool CShadowMap::m_checkValidity()
{
	bool success = true;

	GLuint error = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER);
	switch(error)
	{
	case GL_FRAMEBUFFER_COMPLETE_EXT:
		{
			printf("FrameBufferObjects initialized.");
			break;
		}
	case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
		{
			success = false;
			printf("Incomplete attachment in FBO");
			break;
		}
	case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
		{
			success = false;
			printf("Missing attachment in FBO");
			break;
		}
	case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
		{
			success = false;
			printf("Incomplete dimensions");
			break;
		}
	case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
		{
			success = false;
			printf("Incomplete formats");
			break;
		}
	case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
		{
			success = false;
			printf("Incomplete draw buffer");
			break;
		}
	case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
		{
			success = false;
			printf("Incomplete read buffer");
			break;
		}
	case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
		{
			success = false;
			printf("Framebufferobjects unsupported");
			break;
		}
	}

	return success;
}

void CShadowMap::drawSceneWithShadows()
{
	// for debugging: writes the texture data into an STL vector
	/*int size = m_shadowMapSize[0] * m_shadowMapSize[1] * sizeof(unsigned int);
	unsigned int *bytes = (unsigned int*) malloc( size);
	std::vector<unsigned int> vect;*/

	// 1. pass draw scene as usual

	CScene* pScene = CScene::getScene();
	std::map<int, CShadowData*>::iterator mapIter;

	for(mapIter = m_mapLightAndData.begin(); mapIter != m_mapLightAndData.end(); ++mapIter)
	{
		int lightId = mapIter->first;
		glEnable(GL_LIGHT0+lightId);
	}

	// render the scene with all the lights on
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);	
	pScene->drawMeshes();
	
	for(mapIter = m_mapLightAndData.begin(); mapIter != m_mapLightAndData.end(); ++mapIter)
	{
		int lightId = mapIter->first;
		glDisable(GL_LIGHT0+lightId);
	}


	for(mapIter = m_mapLightAndData.begin(); mapIter != m_mapLightAndData.end(); ++mapIter)
	{
		int lightId = mapIter->first;
		//glEnable(GL_LIGHT0+lightId); // lighting not needed
		CLight* pLight = CLightManager::getLightManager()->getLight(lightId);


		// bind the frame buffer (from here on, we render into an offscreen buffer, until we unbind the frame buffer again)
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_shadowFrameBufferHandle);

		glClearColor(1.0f, 0.0f, 1.0f, 1.0f);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glEnable(GL_CULL_FACE);
		glCullFace(GL_FRONT);

		// save current viewport
		int viewport[4];
		glGetIntegerv(GL_VIEWPORT, (int*) viewport);

		// set viewport and render scene
		glViewport(0,0, m_shadowMapSize[0], m_shadowMapSize[1]);

		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();

		// see also reshape() in viewer.cpp (should be consistent!)
		double znear = 1.0;													// position of near clipping plane
		double zfar = 999.0;													// position of far clipping plane
		double aspectRatio = m_shadowMapSize[0] / m_shadowMapSize[1];	// aspect ratio
		gluPerspective(65, aspectRatio, znear, zfar);

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();

		// get a light's position and look at the center of the scene (we assume this to be (0, 0, 0) !)
		gluLookAt(pLight->m_pos[0], pLight->m_pos[1], pLight->m_pos[2],
			0.0, 0.0, 0.0,
			0.0, 1.0, 0.0);

		// draw the meshes
		pScene->drawMeshes();

		glCullFace(GL_BACK);

		//restore viewport
		glViewport(viewport[0],viewport[1],viewport[2],viewport[3]);

		glMatrixMode(GL_PROJECTION);
		glPopMatrix();

		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();


		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0); // we are back in the original frame buffer
		glClearColor(Globals::clearColor[0], Globals::clearColor[1], Globals::clearColor[2], Globals::clearColor[3]);					// white background
	}

	// render the scene to a quad
	//glEnable(GL_TEXTURE_2D);
	//glActiveTexture(GL_TEXTURE0);
	//glBindTexture(GL_TEXTURE_2D, m_shadowMapColor);
	
	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_NONE);
	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC, GL_ALWAYS);
	
	//glDisable(GL_LIGHTING);
	//glBegin(GL_QUADS);
	//glColor3d(1.0, 1.0, 1.0);
	//glNormal3d(0.0, 1.0, 0.0);
	//glTexCoord2d(0.0, 0.0);		
	//glVertex3d(-2.0, 1.5, 2.0);
	//glTexCoord2d(0.0, 1.0);		
	//glVertex3d(-2.0, 5.5, 2.0);		
	//glTexCoord2d(1.0, 1.0);		
	//glVertex3d(2.0, 5.5, 2.0);
	//glTexCoord2d(1.0, 0.0);		
	//glVertex3d(2.0, 1.5, 2.0);		
	//glEnd();
	//glEnable(GL_LIGHTING);

	//glDisable(GL_TEXTURE_2D);
}
